#include "../config.h"

#include <azur/azur.h>
#include <azur/gint/render.h>

#include "utilities.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fxlibc/printf.h>
#include <sys/types.h>
#include "vector2D.h"
#include <num/num.h>

#include <algorithm>

extern font_t milifont_prop;
extern font_t font_shmup;

void Azur_draw_text(int x, int y, char const *fmt, ...)
{
    char str[128];
    va_list args;
    va_start(args, fmt);
    vsnprintf(str, 128, fmt, args);
    va_end(args);

    dfont( &milifont_prop );

    azrp_text( x, y, C_WHITE, str );
}

void Azur_draw_text_shmup(int x, int y, char const *fmt, ...)
{
    char str[128];
    va_list args;
    va_start(args, fmt);
    vsnprintf(str, 128, fmt, args);
    va_end(args);

    dfont( &font_shmup );

    azrp_text( x+1, y+1, C_RED, str );
    azrp_text( x, y, C_WHITE, str );
}

bool AABB_Collision( SpriteLocator image1, SpriteLocator image2 )
{
    if( (image2.x >= image1.x + image1.image->width) ) return false;
    if( (image2.x + image2.image->width <= image1.x) ) return false;
    if( (image2.y >= image1.y + image1.image->height) ) return false;     
    if( (image2.y + image2.image->height <= image1.y) ) return false;   
    

//    if(    (image2.x >= image1.x + image1.image->width)      
//        || (image2.x + image2.image->width <= image1.x)     
//        || (image2.y >= image1.y + image1.image->height)     
//        || (image2.y + image2.image->height <= image1.y) )   
//        return false;


    // no need to do pixel perfect detection
    return true;
}



bool Pixel_Perfect_Collision( SpriteLocator image1, SpriteLocator image2 )
{
    if( !AABB_Collision( image1, image2 ) )
        return false;                       // bounding boxes not overlapping, impossible to have collision

    /* if we reach that point, this means that we need to perform pixel perfect collsion detection */

    /* First we will calculate the corners of the zone to be tested for collision */
    /* as it is not usefull to check for all the surface of the image, only the */
    /* overlapping area will be verified */



    /* WARNING !! */
    /* P4 format is not supported yet has it is currently focused on use with Azur */
    /* that is mostly P8/RGB565 oriented for ultra fast performances */
    if ( IMAGE_IS_P4(image1.image->format) || IMAGE_IS_P4(image2.image->format) )
        return false;

    int adeltax, adeltay;
    int bdeltax, bdeltay;
    int rows, columns;

    if (image1.x <= image2.x)
    {
        adeltax = image2.x - image1.x;
        columns = MIN(image1.image->width - adeltax, image2.image->width);
        bdeltax = 0;
    }
    else
    {
        bdeltax = image1.x - image2.x;
        columns = MIN(image2.image->width - bdeltax, image1.image->width);
        adeltax = 0;
    }

    if (image1.y <= image2.y)
    {
        adeltay = image2.y - image1.y;
        rows = MIN(image1.image->height - adeltay, image2.image->height);
        bdeltay = 0;
    }
    else
    {
        bdeltay = image1.y - image2.y;
        rows = MIN(image2.image->height - bdeltay, image1.image->height);
        adeltay = 0;
    }

    /* if BOTH images are in a 16bits color format, we will not use masks and will quicken the process */
    if ( IMAGE_IS_RGB16(image1.image->format) && IMAGE_IS_RGB16(image2.image->format) )
    {
        void *data1 = image1.image->data + adeltay * image1.image->stride;
        void *data2 = image2.image->data + bdeltay * image2.image->stride;

        uint16_t *data_u16_1 = (uint16_t *) data1;
        uint16_t *data_u16_2 = (uint16_t *) data2;

        int transp1 = image_alpha(image1.image->format);
        int transp2 = image_alpha(image2.image->format);

        for( int j=0; j<rows; j++)
        {
            for (int i=0; i<columns; i++)
            {
                if(  (data_u16_1[ adeltax + i ] != transp1 )
                  && (data_u16_2[ bdeltax + i ] != transp2 ) )
                  return true;
            }
            data1 += image1.image->stride;
            data_u16_1 = (uint16_t *) data1;
            data2 += image2.image->stride;
            data_u16_2 = (uint16_t *) data2;
        }
        return false;
    }

    /* if BOTH images are in a 8bits color format, we will not use masks and will quicken the process */

    if ( IMAGE_IS_P8(image1.image->format) && IMAGE_IS_P8(image2.image->format) )
    {
        void *data1 = image1.image->data + adeltay * image1.image->stride;
        void *data2 = image2.image->data + bdeltay * image2.image->stride;

        int8_t *data_u8_1 = (int8_t *) data1;
        int8_t *data_u8_2 = (int8_t *) data2;

        int transp1 = image_alpha(image1.image->format);
        int transp2 = image_alpha(image2.image->format);

        for( int j=0; j<rows; j++)
        {
            for (int i=0; i<columns; i++)
            {
                if(  (data_u8_1[ adeltax + i ] != transp1 )
                  && (data_u8_2[ bdeltax + i ] != transp2 ) )
                  return true;
            }
            data1 += image1.image->stride;
            data_u8_1 = (int8_t *) data1;
            data2 += image2.image->stride;
            data_u8_2 = (int8_t *) data2;
        }
        return false;
    }

    /* if we reach that point, this means that images have differnt format and we need to manage this */

    uint8_t d1;
    uint8_t d2;

    void *data1 = image1.image->data + adeltay * image1.image->stride;
    void *data2 = image2.image->data + bdeltay * image2.image->stride;
    
    int8_t *data_u8_1 = (int8_t *) data1;
    uint16_t *data_u16_1 = (uint16_t *) data1;

    int8_t *data_u8_2 = (int8_t *) data2;
    uint16_t *data_u16_2 = (uint16_t *) data2;

    int transp1 = image_alpha(image1.image->format);
    int transp2 = image_alpha(image2.image->format);

    bool im1_P16 = IMAGE_IS_RGB16(image1.image->format);
    bool im1_P8  = IMAGE_IS_P8(image1.image->format);

    bool im2_P16 = IMAGE_IS_RGB16(image2.image->format);
    bool im2_P8  = IMAGE_IS_P8(image2.image->format);

    for( int j=0; j<rows; j++)
    {
        for (int i=0; i<columns; i++)
        {
            /* d1 is set to 1 if pixel of image1 is not transparent and to 0 if transparent */
            /* need to be format dependant here so quite time consumming test at each loop :( )*/
            if (im1_P16)
                d1 = data_u16_1[ adeltax + i ] == transp1 ? 0 : 1;
            else if (im1_P8)
                d1 = data_u8_1[ adeltax + i ] == transp1 ? 0 : 1;
            else d1 = 0;

            /* d2 is set to 1 if pixel of image2 is not transparent and to 0 if transparent */
            /* need to be format dependant here so quite time consumming test at each loop :( )*/
            if (im2_P16)
                d2 = data_u16_2[ bdeltax + i ] == transp2 ? 0 : 1;
            else if (im2_P8)
                d2 = data_u8_2[ bdeltax + i ] == transp2 ? 0 : 1;
            else d2 = 0;

            /* if d1 + d2 = 2 means that both coincident pixels are not transparent and then we have collision*/
            if (d1 + d2 == 2) return true;
        }
        /* we move the pointer to the next line of both images */
        data1 += image1.image->stride;
        data2 += image2.image->stride;
        data_u8_1 = (int8_t *) data1;
        data_u8_2 = (int8_t *) data2;
        data_u16_1 = (uint16_t *) data1;
        data_u16_2 = (uint16_t *) data2;
    }
    return false;
}


/* NEW IMPLEMENTATION - TO TEST */

#include <num/vec.h>
bool get_line_intersection( Vector2D p0_n, Vector2D p1_n, Vector2D p2_n, Vector2D p3_n, Vector2D *pr)
{
    libnum::vec<float,2> p0((float)p0_n.x, (float)p0_n.y);
    libnum::vec<float,2> p1((float)p1_n.x, (float)p1_n.y);
    libnum::vec<float,2> p2((float)p2_n.x, (float)p2_n.y);
    libnum::vec<float,2> p3((float)p3_n.x, (float)p3_n.y);

    libnum::vec<float,2> s1, s2;
    s1 = p1 - p0;
    s2 = p3 - p2;

    float s, t, u;
    u = 1 / (-s2.x * s1.y + s1.x * s2.y);
    s = (-s1.y * (p0.x - p2.x) + s1.x * (p0.y - p2.y)) * u;
    if (s<0) return false;
    if (s>1) return false;

    t = ( s2.x * (p0.y - p2.y) - s2.y * (p0.x - p2.x)) * u; 
    if (t<0) return false;
    if (t>1) return false;

    // Collision detected
    if (pr != NULL)
    {
        (*pr).x = p0.x + (t * s1.x);
        (*pr).y = p0.y + (t * s1.y);
    }
    return true;
}

/* FIRST IMPLEMENTATION USING THE VECTOR2D.H CLASS */

/* O is the origin, dir is the direction of the half line, A and B are the locations of the segment ends*/
bool LineSegment_Collision( Vector2D O, Vector2D dir, Vector2D A, Vector2D B )
{

    Vector2D OA = A - O;

    /* create the point P corresponding to the "end" of the infinite side of the half line */
    /* line length chosen to 1000 units as is it bigger than the diagonal of the screen in pixels*/
    Vector2D P = O.Clone();
    P.Add(dir, libnum::num(1000));

    Vector2D OP = P - O;
    Vector2D OB = B - O;

    /* if the products of determinants is <0, it means A and B are one on each side of the half line, so there is intersection*/
    libnum::num res1 = (OP.Det(OA) * OP.Det(OB));
    
    Vector2D AP = P - A;
    Vector2D AB = B - A;
    Vector2D AO = O - A;

    libnum::num res2 = (AB.Det(AP) * AB.Det(AO));

    if (res1<0 && res2<0) return true;

    return false;
}



bool LineRectangle_Collision( Vector2D Start, Vector2D Direction, Vector2D Min, Vector2D Max )
{   
    Vector2D End = Start.Clone();
    End.Add( Direction, libnum::num(1000));

    /* kind of quick AABB to quickly limit case where accurate check is done */
    if(Start.x < Min.x && End.x < Min.x ) return false;
    if(Start.x > Max.x && End.x > Max.x ) return false;
    if(Start.y < Min.y && End.y < Min.y ) return false;
    if(Start.y > Max.y && End.y > Max.y ) return false;

    Vector2D P1(Min.x, Min.y);
    Vector2D P2(Max.x, Min.y);
    Vector2D P3(Max.x, Max.y);
    Vector2D P4(Min.x, Max.y);

    if (get_line_intersection(Start, End, P1, P2, NULL) ) return true;
    if (get_line_intersection(Start, End, P2, P3, NULL) ) return true;
    if (get_line_intersection(Start, End, P3, P4, NULL) ) return true;
    if (get_line_intersection(Start, End, P4, P1, NULL) ) return true;
    return false;
}



bool LineTriangle_Collision( Vector2D Start, Vector2D Direction, Vector2D A, Vector2D B, Vector2D C )
{
    Vector2D End = Start.Clone();
    End.Add( Direction, libnum::num(1000));

    /* kind of quick AABB to quickly limit case where accurate check is done */
    Vector2D Min( std::min(A.x, std::min(B.x, C.x)), std::min(A.y, std::min(B.y, C.y)) );
    Vector2D Max( std::max(A.x, std::max(B.x, C.x)), std::max(A.y, std::max(B.y, C.y)) );

    if(Start.x < Min.x && End.x < Min.x ) return false;
    if(Start.x > Max.x && End.x > Max.x ) return false;
    if(Start.y < Min.y && End.y < Min.y ) return false;
    if(Start.y > Max.y && End.y > Max.y ) return false;

    if (get_line_intersection(Start, End, A, B, NULL) ) return true;
    if (get_line_intersection(Start, End, B, C, NULL) ) return true;
    if (get_line_intersection(Start, End, C, A, NULL) ) return true;
    return false;
}